#!/bin/bash

##########################################################################
#
# BCKNAME=		# the name of the backup
# BCKROOT=		# destination root dir
# ZFSPOOL=		# ZFS pool name
# BCKS=			# rsync shares
# DSTS=			# destinations (only when using rsync over SSH)
# MAXWEEKS=		# Max. keep weeks
# MAXRUN=		# Max runtime (optional, default:unlimited)
# BWLIMIT=		# Bandwidth limit (optional)
# EXCLUDE=		# Exclude (optional)
# EXCLUDEFROM=		# Exclude from file (optional)
#
# RSYNC_USER=		# Rsync user
# RSYNC_PASSWORD=	# Rsync password
# RSYNC_HOST=		# Rsync host (optional, default:localhost)
# RSYNC_PORT=		# Rsync port
# RSYNC_PROTOCOL=	# Rsync protocol (rsync/ssh, default:rsync)
# RSYNC_ICONV=		# Codepage convert (optional: to,from)
#
# TTIMEOUT=		# Transfer timeout (optional, default:900 sec)
# CTIMEOUT=		# Connect timeout (optional, default:15 sec)
#
# VERBOSE=		# Verbose (optional) 0/1
#
# BACKUP_OPTS=		# any option
#
# COMPRESS=		# new/old, default:old
#
# RSTATUS=		# the remote status rsync "share"
#			# this is required for poll backup
#			# if empty we do the backup immediately
#
##########################################################################

STATUSROOT=/var/spool/backup-remote
LOGROOT=/var/log/backup-remote

RSTATUSROOT=/var/spool/backup-remote-rstatus

BCKDIR=$BCKROOT/$BCKNAME

STATUSDIR=$STATUSROOT/$BCKNAME
RSTATUSDIR=$RSTATUSROOT/$BCKNAME


TODAY=`date +%y%m%d`
TODAYT=`date +%y%m%d_%H%M`

function panic_cfg()
{
	echo "FATAL: Config option missing: $1"
	exit 33
}

function panic_dir()
{
	echo "FATAL: Dir not found: $1"
	exit 34
}

function logme()
{
	MSG="$1"
	LTS=`date '+%T'`
	[ -d $LOGROOT ] || mkdir $LOGROOT
	echo "$LTS $BCKNAME $MSG" >>$LOGROOT/backup-$TODAY.log
	[ -n "$2" ] && echo "$BCKNAME $MSG"
}

function paramcheck() {
	if [ "x$CMDL" != "x" ]; then
		MAXRUN=$CMDL
	fi

	[ "x$BCKNAME" == "x" ] && panic_cfg BCKNAME
	[ "x$BCKROOT" == "x" ] && panic_cfg BCKROOT

	[ "x$ZFSPOOL" == "x" ] && panic_cfg ZFSPOOL

	[ "x$BCKS" == "x" ] && panic_cfg BCKS
	[ "x$MAXWEEKS" == "x" ] && panic_cfg MAXWEEKS

	[ "x$RSYNC_USER" == "x" ] && panic_cfg RSYNC_USER
	[ "x$RSYNC_PASSWORD" == "x" ] && panic_cfg RSYNC_PASSWORD
	[ "x$RSYNC_HOST" == "x" ] && RSYNC_HOST="localhost"
	[ "x$RSYNC_PORT" == "x" ] && panic_cfg RSYNC_PORT

	[ "x$RSYNC_ICONV" != "x" ] && OPT_ICONV="--iconv $RSYNC_ICONV"
	[ "x$BWLIMIT" != "x" ] && OPT_BWLIMIT="--bwlimit $BWLIMIT"
	[ "x$EXCLUDE" != "x" ] && OPT_EXCLUDE="--exclude $EXCLUDE"

	[ "x$RSYNC_PROTOCOL" == "x" ] && RSYNC_PROTOCOL='rsync'

	if [ "x$COMPRESS" == "xnew" ]; then
		OPT_COMPRESS="--compress"
	else
		OPT_COMPRESS="--old-compress"
	fi

	if [ "x$EXCLUDEFROM" != "x" ]; then
		if [ -f $EXCLUDEFROM ]; then
			OPT_EXCLUDEFROM="--exclude-from $EXCLUDEFROM"
		else
			logme "ERROR: $EXCLUDEFROM exclude file doesn't exist!" show
			exit 87
		fi
	fi

	OPT_VERB="-q"
	[ "x$VERBOSE" == "x1" ] && OPT_VERB="-v"

	TTO=900
	[ "x$TTIMEOUT" != "x" ] && TTO=$TTIMEOUT
	CTO=15
	[ "x$CTIMEOUT" != "x" ] && CTO=$CTIMEOUT

	if [ -n "$MAXRUN" ]; then
		limitexec="/usr/bin/timeout $MAXRUN"
	else
		limitexec=
	fi
}

function dircheck() {

	[ -d $BCKROOT ] || panic_dir $BCKROOT

#	[ -d $BCKDIR ] || mkdir $BCKDIR


	if ! [ -d $BCKDIR ]; then
		#btrfs subvolume create $BCKDIR/current
		zfs create ${ZFSPOOL}/${BCKNAME}
	fi

	if ! [ -d $BCKDIR/.zfs/snapshot ]; then
		zfs set snapdir=visible ${ZFSPOOL}
	fi

	[ -d $STATUSROOT ] || mkdir $STATUSROOT
	[ -d $RSTATUSROOT ] || mkdir $RSTATUSROOT
	[ -d $STATUSDIR ] || mkdir $STATUSDIR
	[ -d $RSTATUSDIR ] || mkdir $RSTATUSDIR

#	[ -d $BCKDIR/snapshot ] || mkdir $BCKDIR/snapshot

}


function syncok() {

	errcode=$1

	case "$errcode" in
		30)
			backup_error
			return 30
		;;
		35)
			backup_error
			return 35
		;;
		10)
			backup_error
			return 10
		;;
		11)
			backup_error
			return 11
		;;
		12)
			backup_error
			return 12
		;;
		*)
			return 0
		;;
	esac

}

function syncdone() {

	errcode=$1

	case "$errcode" in
		0)
			return 0
		;;
		24)
			return 0
		;;
		*)
			return 10
		;;
	esac

}

function sync_aborted() {
	errcode=$1
	case "$errcode" in
		20)
			return 20
		;;
		*)
			return 0
		;;
	esac
}

function backup_error() {
	ETS=`date '+%T'`
	logme "  Error during sync at $ETS" show
	sleep 5
	/etc/init.d/stunnel4 restart >/dev/null
	sleep 5
}

function check_remote_status()
{
	if [ "x$RSTATUS" != "x" ]; then

		if [ -f $STATUSDIR/$TODAY.finished ]; then
			return 1
		fi

		#dest
		bsdst=$RSTATUSDIR
		[ -d $bsdst ] || mkdir $bsdst

		case $RSYNC_PROTOCOL in
		    "ssh")
			bssrc="${RSYNC_USER}@${RSYNC_HOST}:/${RSTATUS}"
			/usr/bin/rsync -a -r $OPT_VERB $OPT_BWLIMIT --timeout=900 $OPT_COMPRESS --numeric-ids --delete --delete-excluded --force -e "ssh -p ${RSYNC_PORT}" $bssrc/ $bsdst/ >>$LOGROOT/backup-$TODAY.log 2>&1
			exitcode=$?
		    ;;
		    *)
			bssrc="rsync://${RSYNC_USER}@${RSYNC_HOST}:${RSYNC_PORT}/${RSTATUS}"
			/usr/bin/rsync -a -r $OPT_VERB $OPT_BWLIMIT --timeout=900 --contimeout=15 $OPT_COMPRESS --numeric-ids --delete --delete-excluded --force $bssrc/ $bsdst/ >>$LOGROOT/backup-$TODAY.log 2>&1
			exitcode=$?
		    ;;
		esac

		if [ $exitcode -gt 0 ]; then
			logme "Remote status transfer error, exiting." show
			exit 9
		fi

		if [ -f $bsdst/$TODAY.finished ]; then
			return 0
		fi

		if [ -f $bsdst/$TODAY.error ]; then
			return 0
		fi

		return 1

	else
		return 0
	fi
}


###########################################################

function makebackup() {

	ERR=0

	paramcheck
	dircheck

	# sync
	export RSYNC_PASSWORD

	if check_remote_status; then

		logme "start_backup"

		ERR=0

		n=0
		for BCK in $BCKS; do

			[ -d $BCKDIR/$BCK ] || mkdir $BCKDIR/$BCK

			# make the backup
			for i in 1 2 3 4 5; do
				logme "start_share $BCK"

				case $RSYNC_PROTOCOL in
				    "ssh")
					bsrc="${RSYNC_USER}@${RSYNC_HOST}:/${BCK}"
					if [ "x${DSTS[n]}" != "x" ]; then
					    bdst=$BCKDIR/${DSTS[n]}
					else
					    bdst=$BCKDIR/`basename ${BCK}`
					fi
					$limitexec /usr/bin/rsync -a -r $OPT_VERB -E $OPT_BWLIMIT $OPT_ICONV $OPT_EXCLUDE $OPT_EXCLUDEFROM $BACKUP_OPTS --timeout=$TTO $OPT_COMPRESS --numeric-ids --delete --delete-excluded --force --inplace -e "ssh -p ${RSYNC_PORT}" $bsrc/ $bdst/ >>$LOGROOT/backup-$TODAY.log 2>&1
					SE=$?
				    ;;
				    *)
					bsrc="rsync://${RSYNC_USER}@${RSYNC_HOST}:${RSYNC_PORT}/${BCK}"
					bdst=$BCKDIR/$BCK
					$limitexec /usr/bin/rsync -a -r $OPT_VERB -E $OPT_BWLIMIT $OPT_ICONV $OPT_EXCLUDE $OPT_EXCLUDEFROM $BACKUP_OPTS --timeout=$TTO --contimeout=$CTO $OPT_COMPRESS --numeric-ids --delete --delete-excluded --force --inplace $bsrc/ $bdst/ >>$LOGROOT/backup-$TODAY.log 2>&1
					SE=$?
				    ;;
				esac

				if ! sync_aborted $SE; then
					logme "ABORT, exiting immediately." show
					exit 27
				fi

				if syncok $SE; then break; fi
			done

			if ! syncdone $SE; then
				ERR=1
				logme "end_share $BCK (ERROR)"
			else
				logme "end_share $BCK"
			fi

			# add to index
			n=$((n+1))

		done

		logme "snapshot"

		#btrfs subvolume snapshot -r $BCKDIR $BCKDIR/snapshot/$TODAY >/dev/null
		SSNAME=`date +%y%m%d-%H%M`
		zfs snapshot ${ZFSPOOL}/${BCKNAME}@${SSNAME}

		if [ $ERR -eq 0 ]; then
			touch $STATUSDIR/$TODAY.finished
			[ -f $STATUSDIR/$TODAY.error ] && rm $STATUSDIR/$TODAY.error

			logme "end_backup"
		else
			touch $STATUSDIR/$TODAY.error
			logme "end_backup (ERROR)" show
		fi

	fi

	return $ERR

}

function rotatebackup() {

	paramcheck
	dircheck

	if [ -f $STATUSDIR/$TODAY.finished ]; then

		BSTATUS=`cat $STATUSDIR/$TODAY.finished`

		if [ "$BSTATUS" != "r" ]; then

			logme "start_rotate"

			find $STATUSDIR/ -mindepth 1 -maxdepth 1 -type f -mtime +365

			MAXDAYS=$(( $MAXWEEKS * 7 ))
			zfs list -H -t snapshot -o name -S creation -r ${ZFSPOOL}/${BCKNAME} |tail +${MAXDAYS}

			echo "r" > $STATUSDIR/$TODAY.finished

			logme "end_rotate"
		fi
	fi

	return 0

}

###########################################################
# start
###########################################################

CMDL="$1"

# The originating script name should be backup-* or rotate-*

sn=${0##*/}
func=${sn%%-*}

[ "$func" == "backup" ] && makebackup
[ "$func" == "pbackup" ] && makebackup

[ "$func" == "rotate" ] && rotatebackup
[ "$func" == "protate" ] && rotatebackup

exit 0
